package client;

import java.io.IOException;
import java.net.InetAddress;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.text.Format;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.concurrent.Callable;

import org.apache.log4j.Logger;

import com.sun.grizzly.CallbackHandler;
import com.sun.grizzly.ConnectorHandler;
import com.sun.grizzly.Context;
import com.sun.grizzly.Controller;
import com.sun.grizzly.DefaultProtocolChain;
import com.sun.grizzly.IOEvent;
import com.sun.grizzly.Pipeline;
import com.sun.grizzly.ProtocolChain;
import com.sun.grizzly.ProtocolChainInstanceHandler;
import com.sun.grizzly.ProtocolFilter;
import com.sun.grizzly.TCPSelectorHandler;
import com.sun.grizzly.UDPSelectorHandler;
import com.sun.grizzly.filter.ReadFilter;

public class BtNIOClient {
	private static final Logger	LOG	= Logger.getLogger(BtNIOClient.class);
	private static int debugCounter = 0;
	
	private Controller grizzlyController;
	private Pipeline<Callable> selectorPipeline;
	private ProtocolFilter applicationProtocolFilter;
	
	public void setGrizzlyController(Controller grizzlyController) {
		this.grizzlyController = grizzlyController;
	}
	
	public void setSelectorPipeline(Pipeline<Callable> selectorPipeline) {
		this.selectorPipeline = selectorPipeline;
	}
	
	public void setApplicationProtocolFilter(
			ProtocolFilter applicationProtocolFilter) {
		this.applicationProtocolFilter = applicationProtocolFilter;
	}
	
	/*
	 * selector and connector must both be of the same transport protocol,
	 * either TCP or UDP
	 * 
	 * Note: TCPSelectorHandler is the superclass of UDPSelectorHandler
	 */
	private TCPSelectorHandler selectorHandler;
	
	/**
	 * The default constructor is used when private member initialization is
	 * done using dependency injection or by explicitly calling the setter
	 * methods for the private members.
	 */
	public BtNIOClient() {
		 
	}
	
	public BtNIOClient(Controller grizzlyController)
	{
		this(grizzlyController, null);
	}
	
	public BtNIOClient(Controller grizzlyController, Pipeline<Callable> selectorPipeline)
	{
		this.grizzlyController = grizzlyController;
		this.selectorPipeline = selectorPipeline;
	}
	

	/**
	 * Initialize the grizzly controller, start the controller and initialize
	 * the selectorHandler according to the transport protocol, either TCP or
	 * UDP.
	 * 
	 * @param transportProtocol
	 */
	public void init(Controller.Protocol transportProtocol)
	{
		if(transportProtocol == Controller.Protocol.TCP) {
			selectorHandler = new TCPSelectorHandler(true);
		}
		else if(transportProtocol == Controller.Protocol.UDP) {
			selectorHandler = new UDPSelectorHandler(true);
		}
		
		/*
		 * Set the pipeline to use with the selectorHandler if declared. If not,
		 * it will use the one from the grizzly Controller by default
		 */
		if(selectorPipeline != null)
		{
			selectorHandler.setPipeline(selectorPipeline);
		}
		
		/*
		 * Setting up the protocolChainInstanceHandler for the selectorHandler
		 */
		ProtocolChainInstanceHandler pciHandler = new ClientProtocolChainInstanceHandler();
		ProtocolChain protocolChain = pciHandler.poll();
		/*
		 * We don't need to add to the protocolChain a ReadFilter because a
		 * ParserProtocolFilter extends a ReadFilter 
		 * 
		 * 
		 */
		protocolChain.addFilter(applicationProtocolFilter);
		
		selectorHandler.setProtocolChainInstanceHandler(pciHandler);
		
		grizzlyController.addSelectorHandler(selectorHandler);
	}
	
	public Connector getNewConnector()
	{
		return new Connector();
	}
	
	public class Connector
	{
		private ConnectorHandler<TCPSelectorHandler, ClientCallBackHandler> connectorHandler;
		
		private Connector() {
			
		}
		
		/**
		 * Establishes a TCP/IP Connection to a remoteAddress from the localHost
		 * ethernet interface
		 * 
		 * @param remoteAddress
		 * @throws IOException
		 * @throws InterruptedException 
		 */
		public void connect(SocketAddress remoteAddress) throws IOException
		{
			initConnector();
			connectorHandler.connect(remoteAddress, new ClientCallBackHandler(), selectorHandler);
		}
		
		/**
		 * Establishes a TCP/IP Connection to a remoteAddress from a user defined
		 * ethernet interface specified by a localAddress
		 * 
		 * @param remoteAddress
		 * @param localAddress
		 * @throws IOException
		 * @throws InterruptedException 
		 */
		public void connect(SocketAddress remoteAddress, SocketAddress localAddress) throws IOException
		{
			initConnector();
			connectorHandler.connect(remoteAddress, localAddress, new ClientCallBackHandler(), selectorHandler);
		}
		
		/*
		 * Initialise the connectorHandler. It needs to be done after the
		 * selectorHandler have been initialized AND started
		 */
		private void initConnector()
		{
			/*
			 * Initialize the connector
			 * Requires the controller to have been started before being able to do it.
			 */
			if(connectorHandler == null)
			{
				connectorHandler = selectorHandler.acquireConnectorHandler();
				connectorHandler.setController(grizzlyController);
				System.err.println("connect " + debugCounter++);
			}
		}
		
		public void send(ByteBuffer outputByteBuffer) throws IOException
		{
			/*
			 * Make a blocking write operation
			 */
			long byteWritten = connectorHandler.write(outputByteBuffer, false);
			//System.err.println("Writing "+byteWritten+" bytes");
			outputByteBuffer.clear();
		}
		
		private class ClientCallBackHandler implements CallbackHandler<Context>
		{
			/*
			 * This constructor is added so we can pass the context of the
			 * connection to allow the callbackHandler to log the connection context
			 */
			private final Format dateFormatter = new SimpleDateFormat("yyyy.MM.dd G 'at' HH:mm:ss.SSS z");
			private final Format timeFormatter = new SimpleDateFormat("HH:mm:ss.SSS z");
			
			private InetAddress remoteAddress;
			private InetAddress localAddress;
			private int remotePort;
			private int localPort;
			
			public void onConnect(IOEvent<Context> ioEvent) {
				SelectionKey k = ioEvent.attachment().getSelectionKey();
				try {
					connectorHandler.finishConnect(k);
					remoteAddress = ((SocketChannel)k.channel()).socket().getInetAddress();
					remotePort = ((SocketChannel)k.channel()).socket().getPort();
					localAddress = ((SocketChannel)k.channel()).socket().getLocalAddress();
					localPort = ((SocketChannel)k.channel()).socket().getLocalPort();
					
					if(LOG.isTraceEnabled())
					{
						Date currentDate = new Date(System.currentTimeMillis());
						LOG.trace("connection established on " + dateFormatter.format(currentDate) + " from " + localAddress.getHostAddress() +":" + localPort + " to " + remoteAddress.getHostAddress() + ":" + remotePort);
					}
				} catch (IOException e) {
					e.printStackTrace();
				}
				ioEvent.attachment().getSelectorHandler().register(k, SelectionKey.OP_READ);
			}

			public void onRead(IOEvent<Context> ioEvent) {
				if(LOG.isTraceEnabled())
				{
					Date currentDate = new Date(System.currentTimeMillis());
					LOG.trace("Packet received at " + timeFormatter.format(currentDate) + " from " + remoteAddress.toString() + ":" + remotePort);
				}
				
				try 
				{
					ioEvent.attachment().getProtocolChain().execute(ioEvent.attachment());
				} catch (Exception e) {
					LOG.error("ProtocolChain error while handling received packet from " + remoteAddress.toString() + ":" + remotePort, e);
				}
				
			}

			public void onWrite(IOEvent<Context> ioEvent) {
				if(LOG.isTraceEnabled())
				{
					Date currentDate = new Date(System.currentTimeMillis());
					LOG.trace("Packet sent at " + timeFormatter.format(currentDate) + " to " + remoteAddress.toString() + ":" + remotePort);
				}
			}			
		}
	}
	
	/*
	 * PRIVATE INNER CLASS SECTION
	 */
	
	private class ClientProtocolChainInstanceHandler implements ProtocolChainInstanceHandler
	{
		private final ProtocolChain protocolChain = new DefaultProtocolChain();

		public boolean offer(ProtocolChain instance) {
			return true;
		}

		public ProtocolChain poll() {
			return protocolChain;
		}
	}
}
